//File Description: Create the Pidgin plugin and interface with Pidgin

//Standard pidgin plugin initialization
#define PURPLE_PLUGINS
#ifndef G_GNUC_NULL_TERMINATED
# if __GNUC__ >= 4
#  define G_GNUC_NULL_TERMINATED __attribute__((__sentinel__))
# else
#  define G_GNUC_NULL_TERMINATED
# endif
#endif
#include <notify.h>
#include <version.h>
#include <debug.h> //Debug messages

#include <glib/gprintf.h>
#include <memory.h>

//Plugin path
#define PREF_ROOT "/plugins/gtk/StatusHotKeys"

//Includes needed for pidgin actions
#include <savedstatuses.h> //Manipulate user status
#include <status.h> //User status constants
#include <gtkblist.h> //Get the window for manipulation (pidgin_blist_get_default_gtk_blist)
#include <gtkplugin.h> //PidginPluginUiInfo
#include <gtkconv.h> //Used for OHK_SHOW_PENDING_CONVERSATION

//Interface with other local modules
#include "Environment.h"
#include "Actions.h"
#include "OSInterface.h"
#include "Config.h"

//Set buddy list taskbar visibility
static void plugin_action_toggle_buddy_list_in_taskbar(PurplePluginAction *Action) //Swap current state
{
	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Swapping Buddy List taskbar visibility\n");
	ToggleBuddyListInTaskbar(TaskbarType_Swap); //Toggle the buddy list in the taskbar
}
guint SetTaskbarTimeout_Handle=0;
static gboolean SetTaskbarTimeout(gpointer data) //Set the visibility of the buddy list in the taskbar after a timeout after loading plugin. Otherwise this does not work right in Linux when loading the plugin.
{
	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Setting Buddy List taskbar visibility\n");

	if(SetTaskbarTimeout_Handle) //If this is the actual timeout call, erase the timeout
		gtk_timeout_remove(SetTaskbarTimeout_Handle);

	//Cycle the buddy list taskbar visibility once before setting it finally
	gboolean SetAsHidden=PrefsGetBool("Taskbar_At_Startup");
	ToggleBuddyListInTaskbar(SetAsHidden ? TaskbarType_Visible : TaskbarType_Hidden);
	ToggleBuddyListInTaskbar(SetAsHidden ? TaskbarType_Hidden : TaskbarType_Visible);
	
	return TRUE;
}
static void gtkblist_created_cb(PurpleBuddyList *Blist) //Do some stuff when buddy list window is created
{
	SetTaskbarTimeout_Handle=gtk_timeout_add(500, SetTaskbarTimeout, 0); //Set the taskbar visibility after a timeout
}

//List of actions
static GList* plugin_actions(PurplePlugin *Plugin, gpointer Context)
{
	return g_list_append(NULL, purple_plugin_action_new("Toggle Buddy List in Taskbar", plugin_action_toggle_buddy_list_in_taskbar));
}

//Hook the Status Message change signal to save the last set status (when its not blank)
static void SavedStatusUpdated_cb(PurpleSavedStatus *NewState, PurpleSavedStatus *OldState)
{
	const char* CurrentMessage=purple_savedstatus_get_message(NewState);
	if(CurrentMessage!=NULL && CurrentMessage[0]!=0)
		PrefsSetString("LastUsedStatus", CurrentMessage);
}

//On the input box dialog, when the user "activates" the textbox, have the dialog box press the OK button
static void InputBoxOK(GtkWidget *Entry, GtkWidget *Dialog)
{
	gtk_dialog_response(GTK_DIALOG(Dialog), GTK_RESPONSE_OK);
}

//Process a hot key when it is pressed
void ProcessHotKey(int Identifier)
{
	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Processing hot key\n");

	//Determine which HotKey action was triggered
	gboolean HotKeyProcessed=FALSE; //If the hot key was able to be processed
	for(GList *TraverseActions=Actions;TraverseActions;TraverseActions=g_list_next(TraverseActions))
		if(((HotKeyAction*)TraverseActions->data)->Identifier==Identifier)
		{
			//Process the hot key action
			HotKeyProcessed=TRUE;
			HotKeyAction *Action=(HotKeyAction*)TraverseActions->data;
			if(Action->Type==HKACTIONTYPE_Other) //Other Action
			{
				switch(Action->Info.State)
				{
					case OHK_FOCUS:
						purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Showing Buddy List\n");
						purple_blist_set_visible(TRUE);
						SetBuddyListZOrder(TRUE);
						gdk_window_focus(GetGDKBuddyListWindow(), 0);
						break;
					case OHK_HIDE:
						purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Hiding Buddy List\n");
						purple_blist_set_visible(FALSE);
						break;
					case OHK_BOTTOM:
						purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Moving Buddy List To Bottom\n");
						SetBuddyListZOrder(FALSE);
						break;
					case OHK_TASKBAR_VISIBILITY:
						plugin_action_toggle_buddy_list_in_taskbar(NULL);
						break;
					case OHK_SHOW_PENDING_CONVERSATION:
					{
						GList *PendingConversations=pidgin_conversations_find_unseen_list(PURPLE_CONV_TYPE_IM, PIDGIN_UNSEEN_TEXT, FALSE, 1); //Get the first pending conversation from an IM window
						if(!PendingConversations) //If no pending IM conversations found, try chat windows
							PendingConversations=pidgin_conversations_find_unseen_list(PURPLE_CONV_TYPE_CHAT, PIDGIN_UNSEEN_NICK, FALSE, 1);
						if(!PendingConversations) //If no pending IMs or chat windows found, nothing to do, so exit early
							break;
						pidgin_conv_present_conversation((PurpleConversation*)PendingConversations->data); //Present the window for found conversation
						g_list_free(PendingConversations);
						break;
					}
					default: purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Invalid \"Other Action\" found\n");
				}
				break; //Exit the for loop since action has completed
			}

			//Set the user's Status (HKACTIONTYPE_SetStatus)
			HotKeyAction_Info *StatusHKInfo=&Action->Info;
				//Inform the debug window of the action
				GString *ActionText=GetActionText(Action);
				g_string_prepend(ActionText, "Setting Status: ");
				purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "%s\n", ActionText->str);
				g_string_free(ActionText, TRUE);

				//Determine the new state
				PurpleSavedStatus *CurrentStatus=NULL; //Only pull in the current status state if and when it's needed
				PurpleStatusPrimitive NewState;
				if(StatusHKInfo->State==STATUS_KEEP) //If we are keeping the same status...
				{
					CurrentStatus=purple_savedstatus_get_current();
					NewState=purple_savedstatus_get_type(CurrentStatus);
				}
				else //Change to a new status
				{
					const PurpleStatusPrimitive LocalStateToPidginState[]={0, PURPLE_STATUS_AVAILABLE, PURPLE_STATUS_AWAY, PURPLE_STATUS_UNAVAILABLE, PURPLE_STATUS_INVISIBLE, PURPLE_STATUS_OFFLINE, 0}; //This array must be parallel to enum StatusState
					NewState=LocalStateToPidginState[StatusHKInfo->State];
				}

				//Determine the new message
				char* NewMessage;
				gboolean FreeNewMessage=FALSE; //Whether to free "NewMessage" after it is used or not
				if(StatusHKInfo->MessageState==SSMS_KEEP) //If we are keeping the same message...
				{
					CurrentStatus=(!CurrentStatus ? purple_savedstatus_get_current() : CurrentStatus);
					NewMessage=(char*)purple_savedstatus_get_message(CurrentStatus);
				}
				else if(StatusHKInfo->MessageState==SSMS_STATIC) //Static message
					NewMessage=StatusHKInfo->Message;
				else //SSMS_LAST, SSMS_POPUP, SMSS_POPUP_LAST
				{
					NewMessage=(char*)(StatusHKInfo->MessageState==SSMS_POPUP ? StatusHKInfo->Message : PrefsGetString("LastUsedStatus")); //Different messages for _STATIC and _LAST
					if(StatusHKInfo->MessageState!=SSMS_LAST) //If a popup...
					{
						//Create the dialog box
						GtkWidget *Dialog=gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION, GTK_BUTTONS_OK_CANCEL, "Please enter your new status message");

						//Create the input box to receive the status message
						GtkWidget *InputBox=gtk_entry_new();
						gtk_entry_set_text(GTK_ENTRY(InputBox), NewMessage); //Set the default text
						g_signal_connect(GTK_OBJECT(InputBox), "activate", GTK_SIGNAL_FUNC(InputBoxOK), Dialog); //When the user "activates", have the dialog box press the OK button
						gtk_container_add(GTK_CONTAINER(gtk_container_get_children(GTK_CONTAINER(Dialog))->data), InputBox); //Add to the dialog box
						gtk_widget_show(InputBox); //Make visible

						if(gtk_dialog_run(GTK_DIALOG(Dialog))==GTK_RESPONSE_OK)
						{
							NewMessage=g_strdup(gtk_entry_get_text(GTK_ENTRY(InputBox)));
							FreeNewMessage=TRUE;
							gtk_widget_destroy(Dialog);
						}
						else //Cancel
						{
							gtk_widget_destroy(Dialog);
							break; //If user cancelled, exit prematurely
						}
					}
				}

				//Prepare the new status
				PurpleSavedStatus *NewStatus=purple_savedstatus_new(NULL, NewState); //Prepare the status state
				purple_savedstatus_set_message(NewStatus, NewMessage); //Set the status' message
				purple_savedstatus_activate(NewStatus); //Activate the status

				//Cleanup
				if(FreeNewMessage)
					g_free(NewMessage);

			break; //Exit the for loop since action has completed
		}

	if(!HotKeyProcessed)
		purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Hot key action was not found\n");
}

//Load and unload the plugin
static gboolean plugin_load(PurplePlugin *Plugin)
{
	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Loading\n");

	InitOSInterface();

	//Load the settings
	int NumActions;
	if(PrefExists("Num") && (NumActions=PrefsGetInt("Num"))>0) //If settings are already saved, load from file
		for(int i=0;i<NumActions;i++)
			SaveAction(LoadActionFromPref(i), FALSE);
	else //Load from the default settings
	{
		//Prepare preferences
		purple_prefs_add_none(PREF_ROOT);
		PrefsSetInt("Num", 0);

		//Save default hot keys
		int NumDefaultSettings;
		const HotKeyAction* DefaultSettings=GetDefaultSettings(&NumDefaultSettings);
		for(int i=0;i<NumDefaultSettings;i++)
		{
			HotKeyAction *CopyAction=malloc(sizeof(HotKeyAction));
			memcpy(CopyAction, &DefaultSettings[i], sizeof(HotKeyAction));
			if(CopyAction->Type==HKACTIONTYPE_SetStatus)
				CopyAction->Info.Message=g_strdup(CopyAction->Info.Message);
			SaveAction(CopyAction, TRUE);
		}

		//Save other settings
		PrefsSetString("LastUsedStatus", ""); //The Last Non-Empty Status Message
		PrefsSetBool("Taskbar_At_Startup", TRUE); //Buddy List Taskbar Icon Is Hidden On Startup
	}

	//Status Signal Hooks
	purple_signal_connect(purple_savedstatuses_get_handle(), "savedstatus-changed", Plugin, PURPLE_CALLBACK(SavedStatusUpdated_cb), NULL); //Hook the Status Message change signal to save the last set status (when its not blank)

	//Do stuff when buddy list window is available
	if(GetGDKBuddyListWindow()) //If window is already created, immediately run
		gtkblist_created_cb(NULL);
	else //Otherwise, wait for the signal that it has loaded
		purple_signal_connect(pidgin_blist_get_handle(), "gtkblist-created", Plugin, PURPLE_CALLBACK(gtkblist_created_cb), NULL);

	return TRUE;
}

static gboolean plugin_unload(PurplePlugin *Plugin)
{
	purple_debug(PURPLE_DEBUG_INFO, "Status Hot Keys", "Unloading\n");
	purple_signals_disconnect_by_handle(Plugin); //Unhook signals

	//Remove all actions
	for(int i=g_list_length(Actions)-1;i>=0;i--)
		DeleteAction(i, FALSE);

	CloseOSInterface();

	return TRUE;
}

//Process the plugin's GUI
static GtkWidget* plugin_config_frame(PurplePlugin *Plugin) { return LoadConfigDialog(); }

//Plugin base info
static PidginPluginUiInfo GuiInfo={plugin_config_frame, 0, NULL, NULL, NULL, NULL};
static PurplePluginInfo Info={PURPLE_PLUGIN_MAGIC, PURPLE_MAJOR_VERSION, PURPLE_MINOR_VERSION, PURPLE_PLUGIN_STANDARD, PIDGIN_PLUGIN_TYPE, 0, NULL, PURPLE_PRIORITY_DEFAULT,
	"core-StatusHotKeys", About_ProgramName, About_Version, About_ShortDescription, About_LongDescription, About_Author, About_Website,
	plugin_load, plugin_unload, NULL, &GuiInfo, NULL, NULL, plugin_actions, NULL, NULL, NULL, NULL};

static void init_plugin(PurplePlugin *Plugin) {}
PURPLE_INIT_PLUGIN(StatusHotKeys, init_plugin, Info)

//Get the GDK window for the buddy list
GdkWindow* GetGDKBuddyListWindow(void)
{
	GtkWidget *TheWindow=GetGTKBuddyListWindow();
	return (!TheWindow ? NULL : TheWindow->window);
}

//Get the GTK window for the buddy list
GtkWidget* GetGTKBuddyListWindow(void)
{
	PidginBuddyList *BList=pidgin_blist_get_default_gtk_blist();
	return (BList ? BList->window : NULL);
}

//Set preferences. These are mainly wrappers to purple_prefs_* functions
static const char* GetPrefName(const char* Name) //Get the real name of a preference to retreive
{
	static char PrefBuffer[sizeof(PREF_ROOT)+51];
	g_sprintf(PrefBuffer, "%s/%s", PREF_ROOT, Name);
	return PrefBuffer;
}
static gboolean PrefExistsI(const char* Name) { return purple_prefs_exists(Name); } //Determine if a preference exists (path added [internal])
gboolean PrefExists(const char* Name) { return purple_prefs_exists(GetPrefName(Name)); } //Determine if a preference exists (path not added)
void        PrefsSetBool  (const char* Name, gboolean    Value) { const char* NN=GetPrefName(Name); if(PrefExistsI(NN)) purple_prefs_set_bool  (NN, Value); else purple_prefs_add_bool  (NN, Value); }
void        PrefsSetInt   (const char* Name, int         Value) { const char* NN=GetPrefName(Name); if(PrefExistsI(NN)) purple_prefs_set_int   (NN, Value); else purple_prefs_add_int   (NN, Value); }
void        PrefsSetString(const char* Name, const char* Value) { const char* NN=GetPrefName(Name); if(PrefExistsI(NN)) purple_prefs_set_string(NN, Value); else purple_prefs_add_string(NN, Value); }
gboolean    PrefsGetBool  (const char* Name)                    { return purple_prefs_get_bool  (GetPrefName(Name)); }
int         PrefsGetInt   (const char* Name)                    { return purple_prefs_get_int   (GetPrefName(Name)); }
const char* PrefsGetString(const char* Name)                    { return purple_prefs_get_string(GetPrefName(Name)); }
void        PrefsRemove   (const char* Name)                    {        purple_prefs_remove    (GetPrefName(Name)); }
